// ----------------------------------
// Filename      : Renderer.java
// Created       : 02/11/1999
// Last modified : 02/05/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.engine3d;

import java.awt.*;
import mitrasim.engine3d.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Renderer
{
  public static final boolean kRailroad = true;
  public static final boolean kLine     = false;

  // internal datastructures
  private WVMapping      fWVMapping;
  private Point3D        fViewPoint;
  private SphericalPoint fEyePoint;
  private float[][]      fTransformationMatrix;
  private Graphics       fGr;
  private int            fYOffset;
  private boolean        fPerspectiveMode;

  public Renderer(float wX1, float wY1, float wX2, float wY2,
                  int vX1, int vY1, int vX2, int vY2,
                  Graphics g, int yOffset)
  {
    fWVMapping = new WVMapping(wX1,wY1,wX2,wY2,vX1,vY1,vX2,vY2);
    fViewPoint = new Point3D(0.0f,0.0f,0.0f);
    fEyePoint = new SphericalPoint(1.0f,0.0f,0.0f);
    fTransformationMatrix = new float[3][4];
    for (int i = 0; i < 2; ++i)
      for (int j = 0; j < 3; ++j) {
        fTransformationMatrix[i][j] = 0.0f;
      }
    fPerspectiveMode = true;
    calcTransformationMatrix();
    setGraphics(g);
    setYOffset(yOffset);
  }

  public void setWVMapping(float wX1, float wY1, float wX2, float wY2,
                           int vX1, int vY1, int vX2, int vY2)
  {
    fWVMapping.setWVMapping(wX1,wY1,wX2,wY2,vX1,vY1,vX2,vY2);
  }

  public void setViewPoint(float x, float y, float z)
  {
    fViewPoint.fX = x;
    fViewPoint.fY = y;
    fViewPoint.fZ = z;
    calcTransformationMatrix();
  }

  public void setEyePoint(float rho, float phi, float theta)
  {
    fEyePoint.fRho = rho;
    fEyePoint.fPhi = phi;
    fEyePoint.fTheta = theta;
    calcTransformationMatrix();
  }

  public void setGraphics(Graphics g)
  {
    fGr = g;
  }

  public void setYOffset(int yOffset)
  {
    fYOffset = yOffset;
  }

  public void setPerspectiveMode()
  {
    fPerspectiveMode = true;
  }

  public void setParallelMode()
  {
    fPerspectiveMode = false;
  }

  public float convertRadToDeg(float radians)
  {
    return (float) ((180.0 / Math.PI) * radians);
  }

  public float convertDegToRad(float degrees)
  {
    return (float) ((Math.PI / 180.0) * degrees);
  }

  public Point3D convertSphericalToCartesian(SphericalPoint sphericalPoint)
  {
    Point3D fCartesianPoint = new Point3D();

    fCartesianPoint.fX = (float) (sphericalPoint.fRho *
                                  Math.sin(sphericalPoint.fPhi) *
 	                              Math.cos(sphericalPoint.fTheta));
    fCartesianPoint.fY = (float) (sphericalPoint.fRho *
                                  Math.sin(sphericalPoint.fPhi) *
	                              Math.sin(sphericalPoint.fTheta));
    fCartesianPoint.fZ = (float) (sphericalPoint.fRho *
	                              Math.cos(sphericalPoint.fPhi));

    return fCartesianPoint;
  }

  public SphericalPoint convertCartesianToSpherical(Point3D cartesianPoint)
  {
    SphericalPoint fSphericalPoint = new SphericalPoint();

    fSphericalPoint.fRho = (float) (Math.sqrt(sqr(cartesianPoint.fX) +
                                              sqr(cartesianPoint.fY) +
                                              sqr(cartesianPoint.fZ)));

    if (fSphericalPoint.fRho == 0.0f) {
      fSphericalPoint.fPhi = 0.0f;
      fSphericalPoint.fTheta = 0.0f;
    }
    else {
      fSphericalPoint.fPhi = (float) (Math.acos(cartesianPoint.fZ /
                                                fSphericalPoint.fRho));

      if (cartesianPoint.fX > 0.0f) {
        fSphericalPoint.fTheta = (float) (Math.atan(cartesianPoint.fY /
                                                    cartesianPoint.fX));
      }
      else {
        if (cartesianPoint.fX < 0.0f) {
          fSphericalPoint.fTheta = (float) (Math.atan(cartesianPoint.fY /
	                                                  cartesianPoint.fX) +
                                            Math.PI);
	    }
	    else {
    	  // cartesianPoint.fX == 0.0f
    	  if (cartesianPoint.fY >= 0.0f) {
            fSphericalPoint.fTheta = (float) (Math.PI / 2.0);
    	  }
    	  else {
	        // cartesianPoint.fY <= 0.0f
            fSphericalPoint.fTheta = (float) ((3.0 * Math.PI) / 2.0);
    	  }
    	}
      }
    }

    return fSphericalPoint;
  }

  public void plotPoint(Point3D point)
  {
    ViewportPoint fRenderedPoint = renderPoint(point);
    fGr.fillRect(fRenderedPoint.fX,fRenderedPoint.fY,1,1);
  }

  public void plotDot(Point3D point)
  {
    ViewportPoint fRenderedPoint = renderPoint(point);
    fGr.fillOval(fRenderedPoint.fX - 5,fRenderedPoint.fY - 5,11,11);
    fGr.setColor(Color.white);
    fGr.drawOval(fRenderedPoint.fX - 5,fRenderedPoint.fY - 5,11,11);
  }

  public void plotLine(Point3D point1, Point3D point2)
  {
    ViewportPoint fRenderedPoint1 = renderPoint(point1);
    ViewportPoint fRenderedPoint2 = renderPoint(point2);
    fGr.drawLine(fRenderedPoint1.fX,fRenderedPoint1.fY,
                 fRenderedPoint2.fX,fRenderedPoint2.fY);
  }

  public void plotRailroad(Point3D point1, Point3D point2)
  {
    ViewportPoint fRenderedPoint1 = renderPoint(point1);
    ViewportPoint fRenderedPoint2 = renderPoint(point2);
    fGr.setColor(Color.white);
    fGr.drawLine(fRenderedPoint1.fX - 1,fRenderedPoint1.fY,
                 fRenderedPoint2.fX - 1,fRenderedPoint2.fY);
    fGr.drawLine(fRenderedPoint1.fX + 1,fRenderedPoint1.fY,
                 fRenderedPoint2.fX + 1,fRenderedPoint2.fY);
    fGr.drawLine(fRenderedPoint1.fX,fRenderedPoint1.fY - 1,
                 fRenderedPoint2.fX,fRenderedPoint2.fY - 1);
    fGr.drawLine(fRenderedPoint1.fX,fRenderedPoint1.fY + 1,
                 fRenderedPoint2.fX,fRenderedPoint2.fY + 1);
    fGr.setColor(Color.black);
    fGr.drawLine(fRenderedPoint1.fX,fRenderedPoint1.fY,
                 fRenderedPoint2.fX,fRenderedPoint2.fY);
  }

  public void plotSpline(Point3D point1, Point3D point2,
                         Point3D point3, Point3D point4, int n,
                         boolean Railroad)
  {
    // calculate the product of the Catmull-Rom base-matrix and
    // the 4 controlpoints (that will be interpolated)
    float Cx[] = new float[4];
    float Cy[] = new float[4];
    float Cz[] = new float[4];

    // the x-coordinate
    Cx[0] = (-1 * point1.fX) +
            ( 3 * point2.fX) +
            (-3 * point3.fX) +
            ( 1 * point4.fX);
    Cx[1] = ( 2 * point1.fX) +
            (-5 * point2.fX) +
            ( 4 * point3.fX) +
            (-1 * point4.fX);
    Cx[2] = (-1 * point1.fX) +
            ( 1 * point3.fX);
    Cx[3] = ( 2 * point2.fX);

    // the y-coordinate
    Cy[0] = (-1 * point1.fY) +
            ( 3 * point2.fY) +
            (-3 * point3.fY) +
            ( 1 * point4.fY);
    Cy[1] = ( 2 * point1.fY) +
            (-5 * point2.fY) +
            ( 4 * point3.fY) +
            (-1 * point4.fY);
    Cy[2] = (-1 * point1.fY) +
            ( 1 * point3.fY);
    Cy[3] = ( 2 * point2.fY);

    // the z-coordinate
    Cz[0] = (-1 * point1.fZ) +
            ( 3 * point2.fZ) +
            (-3 * point3.fZ) +
            ( 1 * point4.fZ);
    Cz[1] = ( 2 * point1.fZ) +
            (-5 * point2.fZ) +
            ( 4 * point3.fZ) +
            (-1 * point4.fZ);
    Cz[2] = (-1 * point1.fZ) +
            ( 1 * point3.fZ);
    Cz[3] = ( 2 * point2.fZ);

    for (int j = 0; j < 4; ++j) {
      Cx[j] = (Cx[j] / 2.0f);
      Cy[j] = (Cy[j] / 2.0f);
      Cz[j] = (Cz[j] / 2.0f);
    }

    Point3D Q = new Point3D();
    Point3D PreviousQ = new Point3D();

    // draw the spline-segment between point 2 and point 3
    for (int step = 0; step <= n; ++step) {
      float t = step * (1.0f / n);

      // evaluate using Horner's rule
      Q.fX = Math.round(Cx[3] + t * (Cx[2] + t * (Cx[1] + t * Cx[0])));
      Q.fY = Math.round(Cy[3] + t * (Cy[2] + t * (Cy[1] + t * Cy[0])));
      Q.fZ = Math.round(Cz[3] + t * (Cz[2] + t * (Cz[1] + t * Cz[0])));

      if (step != 0) {
        if (Railroad) {
          plotRailroad(PreviousQ,Q);
        }
        else {
          plotLine(PreviousQ,Q);
        }
      }
      PreviousQ.fX = Q.fX;
      PreviousQ.fY = Q.fY;
      PreviousQ.fZ = Q.fZ;
    }
  }

  public ViewportPoint renderPoint(Point3D point)
  {
    // calculate the coordinates of the point wrt. to the view- and eyepoints
    Point3D fTransformedPoint = transformPointToEye(point);

    Point2D fPerspectivePoint = new Point2D();

    if (fPerspectiveMode == true) {
      // perspective-transform the point
      if (fTransformedPoint.fZ != 0.0f) {
        fPerspectivePoint.fX = fTransformedPoint.fX / fTransformedPoint.fZ;
        fPerspectivePoint.fY = fTransformedPoint.fY / fTransformedPoint.fZ;
      }
      else {
        fPerspectivePoint.fX = fTransformedPoint.fX;
        fPerspectivePoint.fY = fTransformedPoint.fY;
      }
    }
    else {
      // parallel-transform the point
      fPerspectivePoint.fX = fTransformedPoint.fX / fEyePoint.fRho;
      fPerspectivePoint.fY = fTransformedPoint.fY / fEyePoint.fRho;
    }

    // map to screen and adjust for the screen's Y-axis
    ViewportPoint viewportPoint = fWVMapping.map(fPerspectivePoint);
    viewportPoint.fY = fYOffset - viewportPoint.fY;

    return viewportPoint;
  }

  public Point3D addPointByLine(Point3D p1, Point3D p2, float Distance, float Angle)
  {
    // remember the original positions of the line's endpoints
    Point3D Oldp1 = new Point3D(p1.fX,p1.fY,p1.fZ);
    Point3D Oldp2 = new Point3D(p2.fX,p2.fY,p2.fZ);

    Point3D p = new Point3D();
    float RotationAngle = 0.0f;

    // transform the second point
    p2.fX -= p1.fX;
    p2.fY -= p1.fY;

    if (p2.fX == 0.0f) {
      if (p2.fY > 0.0f) {
        RotationAngle = (float) (Math.PI / 2.0);
      }
      else {
        RotationAngle = (float) -(Math.PI / 2.0);
      }
    }
    else {
      RotationAngle = (float) Math.atan(p2.fY / p2.fX);
    }

    // check the direction of the line
    if (Oldp2.fX < Oldp1.fX) {
      Angle += (float) Math.PI;
    }

    p.fX = (float) (Distance * Math.cos(Angle));
    p.fY = (float) (Distance * Math.sin(Angle));

    // now perform the inverse transformation for p
    float x = (float) ((p.fX * Math.cos(RotationAngle)) -
                       (p.fY * Math.sin(RotationAngle)));
    float y = (float) ((p.fX * Math.sin(RotationAngle)) +
                       (p.fY * Math.cos(RotationAngle)));
    p.fX = x + p1.fX;
    p.fY = y + p1.fY;
    p.fZ = p1.fZ;

    // restore the old positions of the line's endpoints
    p1.fX = Oldp1.fX;
    p1.fY = Oldp1.fY;
    p1.fZ = Oldp1.fZ;
    p2.fX = Oldp2.fX;
    p2.fY = Oldp2.fY;
    p2.fZ = Oldp2.fZ;

    return p;
  }

  private Point3D transformPointToEye(Point3D point)
  {
    Point3D fTransformedPoint = new Point3D();

    fTransformedPoint.fX = (fTransformationMatrix[0][0] * point.fX) +
                           (fTransformationMatrix[0][1] * point.fY) +
                            fTransformationMatrix[0][3];
    fTransformedPoint.fY = (fTransformationMatrix[1][0] * point.fX) +
                           (fTransformationMatrix[1][1] * point.fY) +
                           (fTransformationMatrix[1][2] * point.fZ) +
                            fTransformationMatrix[1][3];
    fTransformedPoint.fZ = (fTransformationMatrix[2][0] * point.fX) +
                           (fTransformationMatrix[2][1] * point.fY) +
                           (fTransformationMatrix[2][2] * point.fZ) +
                            fTransformationMatrix[2][3];

    return fTransformedPoint;
  }

  private double sqr(double x)
  {
    return (x * x);
  }

  private void calcTransformationMatrix()
  {
    // pre-calculate some constants
    float fST = (float) Math.sin(fEyePoint.fTheta);
    float fCT = (float) Math.cos(fEyePoint.fTheta);
    float fSP = (float) Math.sin(fEyePoint.fPhi);
    float fCP = (float) Math.cos(fEyePoint.fPhi);

    fTransformationMatrix[0][0] = -fST;
    fTransformationMatrix[0][1] = fCT;
    fTransformationMatrix[0][3] = (fST * fViewPoint.fX) -
                                  (fCT * fViewPoint.fY);

    fTransformationMatrix[1][0] = -fCT * fCP;
    fTransformationMatrix[1][1] = -fST * fCP;
    fTransformationMatrix[1][2] = fSP;
    fTransformationMatrix[1][3] = (fCT * fCP * fViewPoint.fX) +
                                  (fST * fCP * fViewPoint.fY) -
                                  (fSP * fViewPoint.fZ);

    fTransformationMatrix[2][0] = -fCT * fSP;
    fTransformationMatrix[2][1] = -fST * fSP;
    fTransformationMatrix[2][2] = -fCP;
    fTransformationMatrix[2][3] = fEyePoint.fRho +
                                  (fCT * fSP * fViewPoint.fX) +
                                  (fST * fSP * fViewPoint.fY) +
                                  (fCP * fViewPoint.fZ);
  }
}
