// ----------------------------------
// Filename      : Network.java
// Created       : 10/10/1999
// Last modified : 23/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.network;

import java.awt.*;
import java.io.*;
import java.text.*;
import java.util.*;
import mitrasim.exceptions.*;
import mitrasim.network.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Network
{
  public static final boolean kShowTickMarks     = true;
  public static final boolean kDontShowTickMarks = false;

  // the network's internal data-structures
  public float       fLaneWidth;
  public float       fRailroadWidth;
  public String      fTitle;
  public Hashtable   fNodes;
  public Hashtable   fLinks;
  public Vector      fEnvironment;
  public BoundingBox fBoundingBox;
  public long        fTimeStamp;

  // the model's properties
  public  int     fOriginXOffset;
  public  int     fOriginYOffset;
  public  float   fScaleFactor;
  public  float   fZoomFactor;
  public  int     fMapXOffset;
  public  int     fMapYOffset;
  public  String  fMapFilename;
  private boolean fModelSectionRead;

  // the parser for handling the datafile
  private FileParser fNetworkFile;

  // the multilingual interface
  private Messages fMessages;

  public Network(Messages messages)
  {
    fMessages = messages;

    fOriginXOffset = 0;
    fOriginYOffset = 0;
    fScaleFactor = 1.0f;
    fZoomFactor = 1.0f;
    fMapXOffset = 0;
    fMapYOffset = 0;
    fMapFilename = "";
    fModelSectionRead = false;

    clear();
  }

  public void clear()
  {
    fLaneWidth = 1.0f;
    fRailroadWidth = 1.0f;
    fTitle = "";

    // create a hashtable for the nodes and
    // use an initial capacity of 200 nodes and a loadfactor of 0.8
    fNodes = new Hashtable(200,0.8f);

    // create a hashtable for the links and
    // use an initial capacity of 100 links and a loadfactor of 0.8
    fLinks = new Hashtable(100,0.8f);

    fEnvironment = new Vector();

    fBoundingBox = new BoundingBox();

    fTimeStamp = 0;
  }

  public boolean modelSectionRead()
  {
    return fModelSectionRead;
  }

  public void loadFromFile(String filename) throws GeneralError
  {
    // let the fileparser open the file containing the network-description
    fNetworkFile = new FileParser(fMessages,filename);

    // retrieve the network's general settings
    readGeneralSection();

    // retrieve the network-nodes
    readNodesSection();

    // retrieve the network-links
    readLinksSection();

    // retrieve the links' lanechange-permissions
    readLaneChangesSection();

    // retrieve the network-environment
    readEnvironmentSection();

    // retrieve the optional model-section
    readOptionalModelSection();

    // let the fileparser close the datafile
    fNetworkFile.close();
  }

  public Report analyse()
  {
    Report CheckReport = new Report();

    // initialize the links for checking
    Enumeration LinkEnum = fLinks.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) (LinkEnum.nextElement());
      link.fConnectionToMyMainLinkIsAlreadyDefined = false;
      link.fConnectionToMyOnRampIsAlreadyDefined = false;
    }

    // now check the links
    LinkEnum = fLinks.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) (LinkEnum.nextElement());

      // ***************************************************************
      // ** Due to the complexity of lane-change operations and       **
      // ** vehicle transfers between links, I decided not to support **
      // ** the full range of linktypes.                              **
      // **                                                           **
      // ** Only tunnels, highways, merged onramps and                **
      // ** separated offramps are supported.                         **
      // ***************************************************************
      if ((link.fType != Link.kTunnel) &&
          (link.fType != Link.kHighway) &&
          (link.fType != Link.kMergedOnRamp) &&
          (link.fType != Link.kSeparatedOffRamp)) {
        CheckReport.addLine(fMessages.lookup("errorLinkTypeNotSupportedBySimulator",
                                             String.valueOf(link.fID)));
      }

      // there should at least be 2 nodes on the link
      if (link.fNrOfNodes < 2) {
        CheckReport.addLine(fMessages.lookup("errorNotEnoughNodesOnLink",
                                             String.valueOf(link.fID)));
      }

      // the first node must be an origin- or junction-node
      if ((link.fNodes[0].fType != Node.kOrigin) &&
          (link.fNodes[0].fType != Node.kJunction)) {
        CheckReport.addLine(fMessages.lookup("errorInvalidFirstNodeOnLink",
                                             String.valueOf(link.fNodes[0].fID),
                                             link.fNodes[0].translate(fMessages),
                                             String.valueOf(link.fID)));
      }

      // the last node must be a junction- or destination-node
      if ((link.fNodes[link.fNrOfNodes - 1].fType != Node.kJunction) &&
          (link.fNodes[link.fNrOfNodes - 1].fType != Node.kDestination)) {
        CheckReport.addLine(fMessages.lookup("errorInvalidLastNodeOnLink",
                                             String.valueOf(link.fNodes[link.fNrOfNodes - 1].fID),
                                             link.fNodes[link.fNrOfNodes - 1].translate(fMessages),
                                             String.valueOf(link.fID)));
      }

      // all the other nodes must be bend-nodes
      for (int i = 1; i < (link.fNrOfNodes - 1); ++i) {
        if (link.fNodes[i].fType != Node.kBend) {
          CheckReport.addLine(fMessages.lookup("errorInvalidNodeOnLink",
                                               String.valueOf(link.fNodes[i].fID),
                                               link.fNodes[i].translate(fMessages),
                                               String.valueOf(link.fID)));
        }
      }

      // check for duplicate nodes in O(n^2)
      for (int i = 0; i < link.fNrOfNodes; ++i) {
        for (int j = 0; j < i; ++j) {
          if (link.fNodes[i].fID == link.fNodes[j].fID) {
            CheckReport.addLine(fMessages.lookup("errorDuplicateNodeOnLink",
                                                 String.valueOf(link.fNodes[i].fID),
                                                 String.valueOf(link.fID)));
            // don't report the same duplicate node more than once
            break;
          }
        }
      }

      // check for the mininum required number of nodes
      if ((link.fType == Link.kMergedOnRamp) ||
          (link.fType == Link.kSeparatedOffRamp) ||
          (link.fType == Link.kNarrowing) ||
          (link.fType == Link.kWidening)) {
        if (link.fNrOfNodes < 3) {
          CheckReport.addLine(fMessages.lookup("errorIncorrectNrOfNodesOnLink",
                                               String.valueOf(link.fNrOfNodes),
                                               String.valueOf(link.fID)));
        }
      }

      // the number of lanes should be >= 1
      if (link.fNrOfLanes < 1) {
        CheckReport.addLine(fMessages.lookup("errorInvalidNumberOfLanesOnLink",
                                             String.valueOf(link.fID)));
      }

      // the new-number-of-lanes for narrowings should be relevant
      if (link.fType == Link.kNarrowing) {
        if (link.fNewNrOfLanes < 1) {
          CheckReport.addLine(fMessages.lookup("errorInvalidNumberOfLanesOnNarrowing",
                                               String.valueOf(link.fID)));
        }
        else if (link.fNewNrOfLanes >= link.fNrOfLanes) {
          CheckReport.addLine(fMessages.lookup("errorTooMuchLanesForNarrowing",
                                               String.valueOf(link.fID)));
        }
      }

      // the new-number-of-lanes for widenings should be relevant
      if (link.fType == Link.kWidening) {
        if (link.fNewNrOfLanes < 1) {
          CheckReport.addLine(fMessages.lookup("errorInvalidNumberOfLanesOnWidening",
                                               String.valueOf(link.fID)));
        }
        else if (link.fNewNrOfLanes <= link.fNrOfLanes) {
          CheckReport.addLine(fMessages.lookup("errorTooFewLanesForWidening",
                                               String.valueOf(link.fID)));
        }
      }

      // check the link's connectivity
      if (link.fIsTerminator) {
        // a terminator can't be an offramp
        if ((link.fType == Link.kSeparatedOffRamp) ||
            (link.fType == Link.kMergedOffRamp)) {
          CheckReport.addLine(fMessages.lookup("errorTerminatorCantBeOffRamp",
                                               String.valueOf(link.fID)));
        }
        // the last node must be a destination-node
        if (link.fNodes[link.fNrOfNodes - 1].fType != Node.kDestination) {
          CheckReport.addLine(fMessages.lookup("errorTerminatorsLastNodeMustBeDestination",
                                               String.valueOf(link.fID),
                                               String.valueOf(link.fNodes[link.fNrOfNodes - 1].fID),
                                               link.fNodes[link.fNrOfNodes - 1].translate(fMessages)));
        }
      }
      else {
        // the link isn't a terminator so the last node must be a junction-node
        if (link.fNodes[link.fNrOfNodes - 1].fType != Node.kJunction) {
          CheckReport.addLine(fMessages.lookup("errorLinksLastNodeMustBeJunction",
                                               String.valueOf(link.fID),
                                               String.valueOf(link.fNodes[link.fNrOfNodes - 1].fID),
                                               link.fNodes[link.fNrOfNodes - 1].translate(fMessages)));
        }

        // the link connected to this mainlink must exist
        link.fMainLinksNextLink = (Link) (fLinks.get(new Integer(link.fMainLinkConnectedToLink)));
        if (link.fMainLinksNextLink == null) {
          CheckReport.addLine(fMessages.lookup("errorMainLinkConnectionDoesntExist",
                                               String.valueOf(link.fID),
                                               String.valueOf(link.fMainLinkConnectedToLink)));
        }
        else {
          // this link and the next link must join in the same node
          Link NextLink = link.fMainLinksNextLink;
          if (link.fNodes[link.fNrOfNodes - 1].fID != NextLink.fNodes[0].fID) {
            CheckReport.addLine(fMessages.lookup("errorTwoLinksDontJoin",
                                                 String.valueOf(link.fID),
                                                 String.valueOf(NextLink.fID),
                                                 String.valueOf(link.fNodes[link.fNrOfNodes - 1].fID)));
          }

          if (link.fMainLinkConnectedToPart == Link.kConnectedToMainLink) {

            // the number of lanes should be equal
            if (link.fNewNrOfLanes != NextLink.fNrOfLanes) {
              CheckReport.addLine(fMessages.lookup("errorTwoLinksHaveDifferentNrOfLanes",
                                                   String.valueOf(link.fID),
                                                   String.valueOf(NextLink.fID)));
            }

            // there can only exist one connection to the next link's mainlink
            if (NextLink.fConnectionToMyMainLinkIsAlreadyDefined) {
              CheckReport.addLine(fMessages.lookup("errorConnectionFromMainLinkToMainLinkAlreadyDefined",
                                                   String.valueOf(link.fID),
                                                   String.valueOf(NextLink.fID)));
            }
            else {
              NextLink.fConnectionToMyMainLinkIsAlreadyDefined = true;
              NextLink.fMainLinksPreviousLink = link;
            }
          }
          else if (link.fMainLinkConnectedToPart == Link.kConnectedToOnRamp) {

            // the next link should be an onramp
            if ((NextLink.fType != Link.kSeparatedOnRamp) &&
                (NextLink.fType != Link.kMergedOnRamp)) {
              CheckReport.addLine(fMessages.lookup("errorNotAConnectionFromMainLinkToOnRamp",
                                                   String.valueOf(link.fID),
                                                   String.valueOf(NextLink.fID),
                                                   NextLink.translate(fMessages)));
            }
            else {
              // the number of lanes should be equal
              if (link.fNewNrOfLanes != NextLink.fNrOfLanesOnRamp) {
                CheckReport.addLine(fMessages.lookup("errorLinkAndRampHaveDifferentNrOfLanes",
                                                     String.valueOf(link.fID),
                                                     String.valueOf(NextLink.fID)));
              }

              // there can only exist one connection to the next link's onramp
              if (NextLink.fConnectionToMyOnRampIsAlreadyDefined) {
                CheckReport.addLine(fMessages.lookup("errorConnectionFromMainLinkToOnRampAlreadyDefined",
                                                     String.valueOf(link.fID),
                                                     String.valueOf(NextLink.fID)));
              }
              else {
                NextLink.fConnectionToMyOnRampIsAlreadyDefined = true;
                NextLink.fOnRampsPreviousLink = link;
              }
            }
          }
        }

        // check the offramp connectivity (if the link is an offramp)
        if ((link.fType == Link.kSeparatedOffRamp) ||
            (link.fType == Link.kMergedOffRamp)) {
        
          // the link connected to the offramp must exist
          link.fOffRampsNextLink = (Link) (fLinks.get(new Integer(link.fOffRampConnectedToLink)));
          if (link.fOffRampsNextLink == null) {
            CheckReport.addLine(fMessages.lookup("errorOffRampConnectionDoesntExist",
                                                 String.valueOf(link.fID),
                                                 String.valueOf(link.fOffRampConnectedToLink)));
          }
          else {
            // this offramp and the next link must join in the same node
            Link NextLink = link.fOffRampsNextLink;
            if (link.fNodes[link.fNrOfNodes - 1].fID != NextLink.fNodes[0].fID) {
              CheckReport.addLine(fMessages.lookup("errorRampAndLinkDontJoin",
                                                   String.valueOf(link.fID),
                                                   String.valueOf(NextLink.fID),
                                                   String.valueOf(link.fNodes[link.fNrOfNodes - 1].fID)));
            }

            if (link.fOffRampConnectedToPart == Link.kConnectedToMainLink) {

              // the number of lanes should be equal
              if (link.fNrOfLanesOnRamp != NextLink.fNrOfLanes) {
                CheckReport.addLine(fMessages.lookup("errorRampAndLinkHaveDifferentNrOfLanes",
                                                     String.valueOf(link.fID),
                                                     String.valueOf(NextLink.fID)));
              }

              // there can only exist one connection to the next link's mainlink
              if (NextLink.fConnectionToMyMainLinkIsAlreadyDefined) {
                CheckReport.addLine(fMessages.lookup("errorConnectionOffRampToMainLinkAlreadyDefined",
                                                     String.valueOf(link.fID),
                                                     String.valueOf(NextLink.fID)));
              }
              else {
                NextLink.fConnectionToMyMainLinkIsAlreadyDefined = true;
                NextLink.fMainLinksPreviousLink = link;
              }
            }
            else if (link.fOffRampConnectedToPart == Link.kConnectedToOnRamp) {

              // the next link should be an onramp
              if ((NextLink.fType != Link.kSeparatedOnRamp) &&
                  (NextLink.fType != Link.kMergedOnRamp)) {
                CheckReport.addLine(fMessages.lookup("errorNotAConnectionFromOffRampToOnRamp",
                                                     String.valueOf(link.fID),
                                                     String.valueOf(NextLink.fID),
                                                     NextLink.translate(fMessages)));
              }
              else {
                // the number of lanes should be equal
                if (link.fNrOfLanesOnRamp != NextLink.fNrOfLanesOnRamp) {
                  CheckReport.addLine(fMessages.lookup("errorRampAndRampHaveDifferentNrOfLanes",
                                                       String.valueOf(link.fID),
                                                       String.valueOf(NextLink.fID)));
                }

                // there can only exist one connection to the next link's onramp
                if (NextLink.fConnectionToMyOnRampIsAlreadyDefined) {
                  CheckReport.addLine(fMessages.lookup("errorConnectionFromOffRampToOnRampAlreadyDefined",
                                                       String.valueOf(link.fID),
                                                       String.valueOf(NextLink.fID)));
                }
                else {
                  NextLink.fConnectionToMyOnRampIsAlreadyDefined = true;
                  NextLink.fOnRampsPreviousLink = link;
                }
              }
            }
          }
        }
      }
    }

    // rescan the links for detection of illegal network-entrances
    LinkEnum = fLinks.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) (LinkEnum.nextElement());

      // is the link an entrance to the network
      if (link.fMainLinksPreviousLink == null) {

        // a network-entrance must start at an origin-node
         if (link.fNodes[0].fType != Node.kOrigin) {
          CheckReport.addLine(fMessages.lookup("errorIncorrectNodeTypeAtNetworkEntrance",
                                               String.valueOf(link.fNodes[0].fID),
                                               String.valueOf(link.fID)));
        }

        // only tunnels, highways, narrowings and widenings are allowed as network-entrances
        if ((link.fType != Link.kTunnel) &&
            (link.fType != Link.kHighway) &&
            (link.fType != Link.kNarrowing) &&
            (link.fType != Link.kWidening)) {
          CheckReport.addLine(fMessages.lookup("errorIncorrectLinkTypeAtNetworkEntrance",
                                               String.valueOf(link.fID)));
        }
      }
    }

    Enumeration EnvironmentEnum = fEnvironment.elements();
    while (EnvironmentEnum.hasMoreElements()) {
      Environment environment = (Environment) (EnvironmentEnum.nextElement());

      // there should at least be 2 nodes in polygons and railroads
      if ((environment.fType == Environment.kPolygon) ||
          (environment.fType == Environment.kRailroad)) {
        if (environment.fNrOfNodes < 2) {
          CheckReport.addLine(fMessages.lookup("errorNotEnoughNodesInEnvironment",
                                               environment.translate(fMessages)));
        }
      }

      // each node in an environment must be an environment-node
      for (int i = 0; i < environment.fNrOfNodes; ++i) {
        if (environment.fNodes[i].fType != Node.kEnvironment) {
          CheckReport.addLine(fMessages.lookup("errorNoEnvironmentNode",
                                               String.valueOf(environment.fNodes[i].fID),
                                               environment.fNodes[i].translate(fMessages),
                                               environment.translate(fMessages)));
        }
      }
    }

    return CheckReport;
  }

  public void calcGeometry()
  {
    BoundingBox NodesBB = new BoundingBox();
    BoundingBox EnvironmentBB = new BoundingBox();

    // calculate the nodes' bounding-box
    if (fNodes.size() != 0) {

      Enumeration NodeEnum = fNodes.elements();
      boolean First = true;
      while (NodeEnum.hasMoreElements()) {
        Node node = (Node) NodeEnum.nextElement();
        if (First) {
          NodesBB.fMinX = node.fXPos;
          NodesBB.fMinY = node.fYPos;
          NodesBB.fMinZ = node.fZPos;
          NodesBB.fMaxX = node.fXPos;
          NodesBB.fMaxY = node.fYPos;
          NodesBB.fMaxZ = node.fZPos;
          First = false;
        }
        else {
          if (node.fXPos < NodesBB.fMinX) {
            NodesBB.fMinX = node.fXPos;
          }
          if (node.fYPos < NodesBB.fMinY) {
            NodesBB.fMinY = node.fYPos;
          }
          if (node.fZPos < NodesBB.fMinZ) {
            NodesBB.fMinZ = node.fZPos;
          }
          if (node.fXPos > NodesBB.fMaxX) {
            NodesBB.fMaxX = node.fXPos;
          }
          if (node.fYPos > NodesBB.fMaxY) {
            NodesBB.fMaxY = node.fYPos;
          }
          if (node.fZPos > NodesBB.fMaxZ) {
            NodesBB.fMaxZ = node.fZPos;
          }
        }
      }
      NodesBB.fUndefined = false;
    }

    // calculate the environment's bounding-box
    if (fEnvironment.size() != 0) {

      Enumeration EnvironmentEnum = fEnvironment.elements();
      boolean First = true;
      while (EnvironmentEnum.hasMoreElements()) {
        Environment environment = (Environment) EnvironmentEnum.nextElement();
        EnvironmentBB.calcUnion(environment.fBoundingBox);
      }
      EnvironmentBB.fUndefined = false;
    }

    // calculate the union of all the bounding boxes
    fBoundingBox.calcUnion(NodesBB);
    fBoundingBox.calcUnion(EnvironmentBB);

    // calculate each link's geometry
    Enumeration LinkEnum = fLinks.elements();
    while (LinkEnum.hasMoreElements()) {
      Link link = (Link) LinkEnum.nextElement();
      link.calcGeometry();
    }
  }

  private void readGeneralSection() throws GeneralError
  {
    fNetworkFile.readSectionHeader("General");

    fNetworkFile.readNextLineFromFile();
    try {
      fTitle = fNetworkFile.separatePrefix();

      fLaneWidth = (Float.valueOf(fNetworkFile.separatePrefix())).floatValue();

      fRailroadWidth = (Float.valueOf(fNetworkFile.separatePrefix())).floatValue();
    }
    catch (NumberFormatException exc) {
      throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                              GeneralError.kParseError));
    }

    fNetworkFile.readNextLineFromFile();

    fNetworkFile.checkForSectionClosure();
  }

  private void readNodesSection() throws GeneralError
  {
    fNetworkFile.readSectionHeader("Nodes");

    boolean ContinueReading = true;
    do {
      fNetworkFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fNetworkFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          Node node = new Node();

          node.fID = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          node.fXPos = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          node.fYPos = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          node.fZPos = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          String NodeType = fNetworkFile.separatePrefix();
          if (NodeType.equalsIgnoreCase("ORIGIN")) {
            node.fType = Node.kOrigin;
          }
          else if (NodeType.equalsIgnoreCase("DESTINATION")) {
            node.fType = Node.kDestination;
          }
          else if (NodeType.equalsIgnoreCase("JUNCTION")) {
            node.fType = Node.kJunction;
          }
          else if (NodeType.equalsIgnoreCase("BEND")) {
            node.fType = Node.kBend;
          }
          else if (NodeType.equalsIgnoreCase("ENVIRONMENT")) {
            node.fType = Node.kEnvironment;
          }
          else {
            throw (new GeneralError(fMessages.lookup("errorInvalidNodeType",NodeType,String.valueOf(fNetworkFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          node.fName = fNetworkFile.separatePrefix();

          // add the node to the table of nodes
          fNodes.put(new Integer(node.fID),node);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readLinksSection() throws GeneralError
  {
    fNetworkFile.readSectionHeader("Links");

    boolean ContinueReading = true;
    do {
      fNetworkFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fNetworkFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          Link link = new Link();

          link.fID = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          link.fNrOfNodes = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
          if (link.fNrOfNodes < 1) {
            throw (new GeneralError(fMessages.lookup("errorInvalidNrOfNodesSpecified",
                                    String.valueOf(link.fNrOfNodes),String.valueOf(fNetworkFile.fLineNr)),
                                    GeneralError.kNetworkError));
          }

          link.fNodes = new Node[link.fNrOfNodes];
          for (int i = 0; i < link.fNrOfNodes; ++i) {
            int NodeNr = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
            link.fNodes[i] = (Node) fNodes.get(new Integer(NodeNr));
            if (link.fNodes[i] == null) {
              throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",String.valueOf(NodeNr),String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kNetworkError));
            }
          }

          link.fNrOfLanes = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          link.fLengths = new int[link.fNrOfNodes - 1];

          link.fGrades = new float[link.fNrOfNodes - 1];

          link.fSpeedLimit = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          String Connection = fNetworkFile.separatePrefix();
          if (Connection.equalsIgnoreCase("TERMINATOR")) {
            link.fIsTerminator = true;
          }
          else {
            link.fIsTerminator = false;
            // the connection is of the form : link-ID:connection-type
            int ColonPos = Connection.indexOf(':');
            if ((ColonPos > 0) && (ColonPos < (Connection.length() - 1))) {
              link.fMainLinkConnectedToLink = (Integer.valueOf(Connection.substring(0,ColonPos))).intValue();
              String ConnectionType = Connection.substring(ColonPos + 1);
              if (ConnectionType.equalsIgnoreCase("MAINLINK")) {
                link.fMainLinkConnectedToPart = Link.kConnectedToMainLink;
              }
              else if (ConnectionType.equalsIgnoreCase("ONRAMP")) {
                link.fMainLinkConnectedToPart = Link.kConnectedToOnRamp;
              }
              else {
                throw (new GeneralError(fMessages.lookup("errorInvalidConnection",ConnectionType,String.valueOf(fNetworkFile.fLineNr)),
                                        GeneralError.kParseError));
              }
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidConnection",Connection,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }
          }

          String LinkType = fNetworkFile.separatePrefix();
          if (LinkType.equalsIgnoreCase("TUNNEL")) {
            link.fType = Link.kTunnel;
          }
          else if (LinkType.equalsIgnoreCase("HIGHWAY")) {
            link.fType = Link.kHighway;
          }
          else if (LinkType.equalsIgnoreCase("ONRAMP")) {
            link.fType = Link.kSeparatedOnRamp;
          }
          else if (LinkType.equalsIgnoreCase("OFFRAMP")) {
            link.fType = Link.kSeparatedOffRamp;
          }
          else if (LinkType.equalsIgnoreCase("NARROWING")) {
            link.fType = Link.kNarrowing;
          }
          else if (LinkType.equalsIgnoreCase("WIDENING")) {
            link.fType = Link.kWidening;
          }
          else {
            throw (new GeneralError(fMessages.lookup("errorInvalidLinkType",LinkType,String.valueOf(fNetworkFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          // retrieve the ramp's type (separated or merged)
          if ((link.fType == Link.kSeparatedOnRamp) ||
              (link.fType == Link.kSeparatedOffRamp)) {
            String RampType = fNetworkFile.separatePrefix();
            if (RampType.equalsIgnoreCase("SEPARATED")) {
              // ramptype is already separated
            }
            else if (RampType.equalsIgnoreCase("MERGED")) {
              if (link.fType == Link.kSeparatedOnRamp) {
                link.fType = Link.kMergedOnRamp;
              }
              else if (link.fType == Link.kSeparatedOffRamp) {
                link.fType = Link.kMergedOffRamp;
              }
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidRampType",RampType,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }
          }

          // determine the new number of lanes
          if ((link.fType == Link.kTunnel) ||
              (link.fType == Link.kHighway)) {
            link.fNewNrOfLanes = link.fNrOfLanes;
          }
          else {
            int NrOfLanes = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

            if (link.fType == Link.kSeparatedOnRamp) {
              link.fNrOfLanesOnRamp = NrOfLanes;
              link.fNewNrOfLanes = link.fNrOfLanes + link.fNrOfLanesOnRamp;
            }
            else if (link.fType == Link.kMergedOnRamp) {
              link.fNrOfLanesOnRamp = NrOfLanes;
              link.fNewNrOfLanes = link.fNrOfLanes;
            }
            else if (link.fType == Link.kSeparatedOffRamp) {
              link.fNrOfLanesOnRamp = NrOfLanes;
              link.fNewNrOfLanes = link.fNrOfLanes;
            }
            else if (link.fType == Link.kMergedOffRamp) {
              link.fNrOfLanesOnRamp = NrOfLanes;
              link.fNewNrOfLanes = link.fNrOfLanes - link.fNrOfLanesOnRamp;
            }
            else if (link.fType == Link.kNarrowing) {
              link.fNewNrOfLanes = NrOfLanes;
            }
            else if (link.fType == Link.kWidening) {
              link.fNewNrOfLanes = NrOfLanes;
            }
          }

          // retrieve the location for ramps, narrowings and widenings
          if ((link.fType == Link.kSeparatedOnRamp) ||
              (link.fType == Link.kMergedOnRamp) ||
              (link.fType == Link.kSeparatedOffRamp) ||
              (link.fType == Link.kMergedOffRamp) ||
              (link.fType == Link.kNarrowing) ||
              (link.fType == Link.kWidening)) {
            String Location = fNetworkFile.separatePrefix();
            if (Location.equalsIgnoreCase("LEFT")) {
              link.fLocation = Link.kLeft;
            }
            else if (Location.equalsIgnoreCase("RIGHT")) {
              link.fLocation = Link.kRight;
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidLocation",Location,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }
          }

          // retrieve the connection for the offramps
          if ((link.fType == Link.kSeparatedOffRamp) ||
              (link.fType == Link.kMergedOffRamp)) {

            Connection = fNetworkFile.separatePrefix();
            // the connection is of the form : link-ID:connection-type
            int ColonPos = Connection.indexOf(':');
            if ((ColonPos > 0) && (ColonPos < (Connection.length() - 1))) {
              link.fOffRampConnectedToLink = (Integer.valueOf(Connection.substring(0,ColonPos))).intValue();
              String ConnectionType = Connection.substring(ColonPos + 1);
              if (ConnectionType.equalsIgnoreCase("MAINLINK")) {
                link.fOffRampConnectedToPart = Link.kConnectedToMainLink;
              }
              else if (ConnectionType.equalsIgnoreCase("ONRAMP")) {
                link.fOffRampConnectedToPart = Link.kConnectedToOnRamp;
              }
              else {
                throw (new GeneralError(fMessages.lookup("errorInvalidConnection",ConnectionType,String.valueOf(fNetworkFile.fLineNr)),
                                        GeneralError.kParseError));
              }
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidConnection",Connection,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }
          }

          link.createDefaultLaneChangePermissions();

          link.fName = fNetworkFile.separatePrefix();

          // add the link to the table of links
          fLinks.put(new Integer(link.fID),link);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readLaneChangesSection() throws GeneralError
  {
    fNetworkFile.readSectionHeader("Lanechanges");

    boolean ContinueReading = true;
    do {
      fNetworkFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fNetworkFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          int LinkID = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

          Link link = (Link) fLinks.get(new Integer(LinkID));
          if (link == null) {
            throw (new GeneralError(fMessages.lookup("errorInvalidLinkReferenced",
                                    String.valueOf(LinkID),String.valueOf(fNetworkFile.fLineNr)),
                                    GeneralError.kNetworkError));
          }

          for (int i = 0; i < link.fNrOfLaneChangePermissions; ++i) {
            char LaneChangePermission = (fNetworkFile.separatePrefix()).charAt(0);
            if ((LaneChangePermission == 'y') ||
                (LaneChangePermission == 'Y')) {
              link.fLaneChangePermissions[i] = Link.kAllowed;
            }
            else if ((LaneChangePermission == 'n') ||
                     (LaneChangePermission == 'N')) {
              link.fLaneChangePermissions[i] = Link.kNotAllowed;
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidLaneChangePermissionSpecified",
                                      String.valueOf(LaneChangePermission),String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kNetworkError));
            }
          }
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readEnvironmentSection() throws GeneralError
  {
    fNetworkFile.readSectionHeader("Environment");

    boolean ContinueReading = true;
    do {
      fNetworkFile.readNextLineFromFile();

      // have we reached the end of this section ?
      if (fNetworkFile.fLineRead.charAt(0) == '}') {
        ContinueReading = false;
      }
      else {
        // parse the line
        try {
          Environment environment = new Environment();

          String EnvironmentType = fNetworkFile.separatePrefix();
          if (EnvironmentType.equalsIgnoreCase("RECTANGLE")) {
            environment.fType = Environment.kRectangle;
          }
          else if (EnvironmentType.equalsIgnoreCase("POLYGON")) {
            environment.fType = Environment.kPolygon;
          }
          else if (EnvironmentType.equalsIgnoreCase("OVAL")) {
            environment.fType = Environment.kOval;
          }
          else if (EnvironmentType.equalsIgnoreCase("RAILROAD")) {
            environment.fType = Environment.kRailroad;
          }
          else {
            throw (new GeneralError(fMessages.lookup("errorInvalidEnvironmentType",EnvironmentType,String.valueOf(fNetworkFile.fLineNr)),
                                    GeneralError.kParseError));
          }

          // retrieve the color and filltype for rectangles, polygons and ovals
          if ((environment.fType == Environment.kRectangle) ||
              (environment.fType == Environment.kPolygon) ||
              (environment.fType == Environment.kOval)) {
            String EnvironmentColor = fNetworkFile.separatePrefix();
            if (EnvironmentColor.equalsIgnoreCase("BLACK")) {
              environment.fColor = Color.black;
            }
            else if (EnvironmentColor.equalsIgnoreCase("BLUE")) {
              environment.fColor = Color.blue;
            }
            else if (EnvironmentColor.equalsIgnoreCase("CYAN")) {
              environment.fColor = Color.cyan;
            }
            else if (EnvironmentColor.equalsIgnoreCase("GRAY")) {
              environment.fColor = Color.gray;
            }
            else if (EnvironmentColor.equalsIgnoreCase("GREEN")) {
              environment.fColor = Color.green;
            }
            else if (EnvironmentColor.equalsIgnoreCase("MAGENTA")) {
              environment.fColor = Color.magenta;
            }
            else if (EnvironmentColor.equalsIgnoreCase("ORANGE")) {
              environment.fColor = Color.orange;
            }
            else if (EnvironmentColor.equalsIgnoreCase("RED")) {
              environment.fColor = Color.red;
            }
            else if (EnvironmentColor.equalsIgnoreCase("WHITE")) {
              environment.fColor = Color.white;
            }
            else if (EnvironmentColor.equalsIgnoreCase("YELLOW")) {
              environment.fColor = Color.yellow;
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidEnvironmentColor",EnvironmentColor,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }

            String EnvironmentFill = fNetworkFile.separatePrefix();
            if (EnvironmentFill.equalsIgnoreCase("OUTLINED")) {
              environment.fFilled = false;
            }
            else if (EnvironmentFill.equalsIgnoreCase("FILLED")) {
              environment.fFilled = true;
            }
            else {
              throw (new GeneralError(fMessages.lookup("errorInvalidEnvironmentFill",EnvironmentFill,String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kParseError));
            }
          }

          // retrieve all the environment-nodes
          if ((environment.fType == Environment.kRectangle) ||
              (environment.fType == Environment.kPolygon) ||
              (environment.fType == Environment.kRailroad)) {
            if (environment.fType == Environment.kRectangle) {
              environment.fNrOfNodes = 2;
            }
            else {
              environment.fNrOfNodes = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
            }
            if (environment.fNrOfNodes < 1) {
              throw (new GeneralError(fMessages.lookup("errorInvalidNrOfNodesSpecified",
                                      String.valueOf(environment.fNrOfNodes),String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kNetworkError));
            }
            environment.fNodes = new Node[environment.fNrOfNodes];
            for (int i = 0; i < environment.fNrOfNodes; ++i) {
              int NodeNr = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
              environment.fNodes[i] = (Node) fNodes.get(new Integer(NodeNr));
              if (environment.fNodes[i] == null) {
                throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",String.valueOf(NodeNr),String.valueOf(fNetworkFile.fLineNr)),
                                        GeneralError.kNetworkError));
              }
            }
          }
          else if (environment.fType == Environment.kOval) {
            environment.fNrOfNodes = 1;
            environment.fNodes = new Node[environment.fNrOfNodes];
            int NodeNr = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
            environment.fNodes[0] = (Node) fNodes.get(new Integer(NodeNr));
            if (environment.fNodes[0] == null) {
              throw (new GeneralError(fMessages.lookup("errorInvalidNodeReferenced",String.valueOf(NodeNr),String.valueOf(fNetworkFile.fLineNr)),
                                      GeneralError.kNetworkError));
            }
            environment.fXRadius = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
            environment.fYRadius = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();
          }

          environment.fName = fNetworkFile.separatePrefix();

          // calculate the environment's bounding-box
          environment.calcBoundingBox();

          // add the environment to the network's environment
          fEnvironment.addElement(environment);
        }
        catch (NumberFormatException exc) {
          throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                                  GeneralError.kParseError));
        }
      }
    }
    while (ContinueReading);
  }

  private void readOptionalModelSection() throws GeneralError
  {
    fModelSectionRead = false;

    try {
      fNetworkFile.readSectionHeader("Model");
    }
    catch (GeneralError err) {
      // no 'model'-section was found
      return;
    }

    fNetworkFile.readNextLineFromFile();
    try {
      fOriginXOffset = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

      fOriginYOffset = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

      fScaleFactor = (Float.valueOf(fNetworkFile.separatePrefix())).floatValue();

      fZoomFactor = (Float.valueOf(fNetworkFile.separatePrefix())).floatValue();

      fMapXOffset = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

      fMapYOffset = (Integer.valueOf(fNetworkFile.separatePrefix())).intValue();

      fMapFilename = fNetworkFile.separatePrefix();
    }
    catch (NumberFormatException exc) {
      throw (new GeneralError(fMessages.lookup("errorInvalidNumber",String.valueOf(fNetworkFile.fLineNr)),
                              GeneralError.kParseError));
    }

    fNetworkFile.readNextLineFromFile();
    // does the section-block ends with '}' ?
    if (fNetworkFile.fLineRead.charAt(0) != '}') {
      throw (new GeneralError(fMessages.lookup("errorCloseBraceExpected",String.valueOf(fNetworkFile.fLineNr)),
                              GeneralError.kParseError));
    }

    fModelSectionRead = true;
  }

  public boolean saveToFile(String Filename)
  {
    try {
      FileOutputStream NetworkFile = new FileOutputStream(Filename);
      PrintWriter NetworkWriter = new PrintWriter(NetworkFile);

      final int kIndexWidth         = 5 + 1;
      final int kCoordinateWidth    = 7 + 1;
      final int kNrOfLaneLanesWidth = 2 + 1;
      final int kSpeedLimit         = 3 + 1;

      String HorizontalLine = "# ----------------------------------";

      Date CurrentDate = new Date();
      String FilenameLine = "# Filename : " + Filename;
      String CreatedLine = "# Created  : " +
                           DateFormat.getDateInstance(DateFormat.FULL,Locale.ENGLISH).format(CurrentDate);

      int FilenameLineLength = FilenameLine.length();
      int CreatedLineLength = CreatedLine.length();
      int HorizontalLineLength = HorizontalLine.length();

      int MaxLineLength = HorizontalLineLength;
      if (FilenameLineLength > MaxLineLength) {
        MaxLineLength = FilenameLineLength;
      }
      if (CreatedLineLength > MaxLineLength) {
        MaxLineLength = CreatedLineLength;
      }

      if (MaxLineLength > HorizontalLineLength) {
        for (int i = 1; i <= (MaxLineLength - HorizontalLineLength); ++i) {
          HorizontalLine += "-";
        }
      }

      NetworkWriter.println(HorizontalLine);
      NetworkWriter.println("# Filename : " + Filename);
      NetworkWriter.println(CreatedLine);
      NetworkWriter.println("# Author   : Sven Maerivoet");
      NetworkWriter.println("# Target   : Mitrasim 2000");
      NetworkWriter.println("#");
      NetworkWriter.println("# All rights reserved.");
      NetworkWriter.println("# Universitaire Instelling Antwerpen");
      NetworkWriter.println(HorizontalLine);
      NetworkWriter.println();
      NetworkWriter.println("# Note : if the data doesn't fit on a line then terminate the line");
      NetworkWriter.println("#        with a backslash \\ and continue on the next line (strings can't be split).");
      NetworkWriter.println();

      // save the general-section
      NetworkWriter.println("[General]");
      NetworkWriter.println("# This section is defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   networktitle  lanewidth  railroadwidth");
      NetworkWriter.println("#");
      NetworkWriter.println("# where lanewidth is expressed in meters (float).");
      NetworkWriter.println("{");
      NetworkWriter.println("  \"" +
                            fTitle +
                            "\"  " +
                            String.valueOf(fLaneWidth) +
                            "  " +
                            String.valueOf(fRailroadWidth));
      NetworkWriter.println("}");

      NetworkWriter.println();
      // save the nodes-section
      NetworkWriter.println("[Nodes]");
      NetworkWriter.println("# Each node is defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   ID  xpos  ypos  zpos  nodetype  [nodename]");
      NetworkWriter.println("#");
      NetworkWriter.println("# where xpos, ypos and zpos are expressed in meters (integers).");
      NetworkWriter.println("#");
      NetworkWriter.println("# Available node-types are : origin, destination, junction, bend, environment.");
      NetworkWriter.println("{");
      Enumeration NodeEnum = fNodes.elements();
      while (NodeEnum.hasMoreElements()) {
        Node node = (Node) NodeEnum.nextElement();

        NetworkWriter.print("  ");
        NetworkWriter.print(expand(String.valueOf(node.fID),kIndexWidth));
        NetworkWriter.print("  ");
        if (node.fXPos < 0) {
          NetworkWriter.print(expand(String.valueOf(node.fXPos),kCoordinateWidth));
        }
        else {
          NetworkWriter.print(" ");
          NetworkWriter.print(expand(String.valueOf(node.fXPos),kCoordinateWidth - 1));
        }
        NetworkWriter.print("  ");
        if (node.fYPos < 0) {
          NetworkWriter.print(expand(String.valueOf(node.fYPos),kCoordinateWidth));
        }
        else {
          NetworkWriter.print(" ");
          NetworkWriter.print(expand(String.valueOf(node.fYPos),kCoordinateWidth - 1));
        }
        NetworkWriter.print("  ");
        if (node.fZPos < 0) {
          NetworkWriter.print(expand(String.valueOf(node.fZPos),kCoordinateWidth));
        }
        else {
          NetworkWriter.print(" ");
          NetworkWriter.print(expand(String.valueOf(node.fZPos),kCoordinateWidth - 1));
        }
        NetworkWriter.print("  ");
        if (node.fType == Node.kOrigin) {
          NetworkWriter.print(expand("origin",12));
        }
        else if (node.fType == Node.kDestination) {
          NetworkWriter.print(expand("destination",12));
        }
        else if (node.fType == Node.kJunction) {
          NetworkWriter.print(expand("junction",12));
        }
        else if (node.fType == Node.kBend) {
          NetworkWriter.print(expand("bend",12));
        }
        else if (node.fType == Node.kEnvironment) {
          NetworkWriter.print(expand("environment",12));
        }
        if (!node.fName.equalsIgnoreCase("")) {
          NetworkWriter.print("  ");
          NetworkWriter.print("\"" + node.fName + "\"");
        }
        NetworkWriter.println();
      }
      NetworkWriter.println("}");

      NetworkWriter.println();
      // save the links-section
      NetworkWriter.println("[Links]");
      NetworkWriter.println("# Each link is defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   ID  #nodes  nodes  #lanes  speedlimit  linkconnection  linktype  parameters  [linkname]");
      NetworkWriter.println("#");
      NetworkWriter.println("# where speedlimit is expressed in km/h (integer) and linkconnection is of the form :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   link-ID:connection-type");
      NetworkWriter.println("#");
      NetworkWriter.println("#     or if the link is at the edge of the network :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   terminator");
      NetworkWriter.println("#");
      NetworkWriter.println("# Available connection-types are : mainlink, onramp.");
      NetworkWriter.println("# Following link-types (with their parameters) are supported :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   tunnel    (no extra parameters needed)");
      NetworkWriter.println("#   highway   (no extra parameters needed)");
      NetworkWriter.println("#   onramp    (parameters : ramptype  number-of-lanes-on-ramp  location)");
      NetworkWriter.println("#   offramp   (parameters : ramptype  number-of-lanes-on-ramp  location  link-ID:connection-type)");
      NetworkWriter.println("#   narrowing (parameters : new-number-of-lanes  location)");
      NetworkWriter.println("#   widening  (parameters : new-number-of-lanes  location)");
      NetworkWriter.println("#");
      NetworkWriter.println("# Available ramptypes are : separated, merged.");
      NetworkWriter.println("# Available locations are : left, right.");
      NetworkWriter.println("{");
      // scan for the largest nodes-array
      int MaxNrOfNodes = 0;
      Enumeration LinkEnum = fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        if (link.fNrOfNodes > MaxNrOfNodes) {
          MaxNrOfNodes = link.fNrOfNodes;
        }
      }
      LinkEnum = fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        NetworkWriter.print("  ");
        NetworkWriter.print(expand(String.valueOf(link.fID),kIndexWidth));
        NetworkWriter.print("  ");
        NetworkWriter.print(expand(String.valueOf(link.fNrOfNodes),kIndexWidth));
        NetworkWriter.print("  ");
        String Nodes = "";
        for (int i = 0; i < link.fNrOfNodes; ++i) {
          Nodes = Nodes + expand(String.valueOf(link.fNodes[i].fID),kIndexWidth) + "  ";
        }
        // align all the nodes of the different links
        NetworkWriter.print(expand(Nodes,(MaxNrOfNodes - 1) * (kIndexWidth + 2)));
        NetworkWriter.print(expand(String.valueOf(link.fNrOfLanes),kNrOfLaneLanesWidth));
        NetworkWriter.print("  ");
        NetworkWriter.print(expand(String.valueOf(link.fSpeedLimit),kSpeedLimit));
        NetworkWriter.print("  ");
        if (link.fIsTerminator) {
          NetworkWriter.print(expand("terminator",9 + kIndexWidth));
        }
        else {
          String Connection = String.valueOf(link.fMainLinkConnectedToLink) + ":";
          if (link.fMainLinkConnectedToPart == Link.kConnectedToMainLink) {
            Connection += "mainlink";
          }
          else {
            Connection += "onramp";
          }
          NetworkWriter.print(expand(Connection,9 + kIndexWidth));
        }
        if (link.fType == Link.kTunnel) {
          NetworkWriter.print(expand("tunnel",10));
        }
        else if (link.fType == Link.kHighway) {
          NetworkWriter.print(expand("highway",10));
        }
        else {
          if ((link.fType == Link.kSeparatedOnRamp) ||
              (link.fType == Link.kMergedOnRamp)) {
            NetworkWriter.print(expand("onramp",10));
          }
          else if ((link.fType == Link.kSeparatedOffRamp) ||
                   (link.fType == Link.kMergedOffRamp)) {
            NetworkWriter.print(expand("offramp",10));
          }
          else if (link.fType == Link.kNarrowing) {
            NetworkWriter.print(expand("narrowing",10));
          }
          else if (link.fType == Link.kWidening) {
            NetworkWriter.print(expand("widening",10));
          }
          NetworkWriter.print("  ");
          if ((link.fType == Link.kSeparatedOnRamp) ||
              (link.fType == Link.kSeparatedOffRamp)) {
            NetworkWriter.print(expand("separated",10));
            NetworkWriter.print("  ");
          }
          else if ((link.fType == Link.kMergedOnRamp) ||
                   (link.fType == Link.kMergedOffRamp)) {
            NetworkWriter.print(expand("merged",10));
            NetworkWriter.print("  ");
          }
          if ((link.fType == Link.kSeparatedOnRamp) ||
              (link.fType == Link.kSeparatedOffRamp) ||
              (link.fType == Link.kMergedOnRamp) ||
                   (link.fType == Link.kMergedOffRamp)) {
            NetworkWriter.print(expand(String.valueOf(link.fNrOfLanesOnRamp),kNrOfLaneLanesWidth));
            NetworkWriter.print("  ");
          }
          else {
            NetworkWriter.print(expand(String.valueOf(link.fNewNrOfLanes),kNrOfLaneLanesWidth));
            NetworkWriter.print("  ");
          }
          if (link.fLocation == Link.kLeft) {
            NetworkWriter.print(expand("left",6));
          }
          else {
            NetworkWriter.print(expand("right",6));
          }
          NetworkWriter.print("  ");
          if ((link.fType == Link.kSeparatedOffRamp) ||
              (link.fType == Link.kMergedOffRamp)) {
            String Connection = String.valueOf(link.fOffRampConnectedToLink) + ":";
            if (link.fOffRampConnectedToPart == Link.kConnectedToMainLink) {
              Connection += "mainlink";
            }
            else {
              Connection += "onramp";
            }
            NetworkWriter.print(expand(Connection,9 + kIndexWidth));              
          }          
        }
        if (!link.fName.equalsIgnoreCase("")) {
          NetworkWriter.print("  ");
          NetworkWriter.print("\"" + link.fName + "\"");
        }
        NetworkWriter.println();
      }
      NetworkWriter.println("}");

      NetworkWriter.println();
      // save the lanechanges-section
      NetworkWriter.println("[Lanechanges]");
      NetworkWriter.println("# Each link's lanechange-permissions are defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   link-ID  permissionlist");
      NetworkWriter.println("#");
      NetworkWriter.println("# where each element in the permissionlist is a couple of the form :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   lane i -> lane i + 1    lane i + 1 -> lane i");
      NetworkWriter.println("#");
      NetworkWriter.println("# where i starts at 1 and each element in a couple is of the form :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   Y if the lanechange is allowed,");
      NetworkWriter.println("#   N if the lanechange is not allowed.");
      NetworkWriter.println("{");
      LinkEnum = fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        NetworkWriter.print("  ");
        NetworkWriter.print(expand(String.valueOf(link.fID),kIndexWidth));
        NetworkWriter.print("  ");
        for (int i = 0; i < link.fNrOfLaneChangePermissions; ++i) {
          if (link.fLaneChangePermissions[i] == Link.kAllowed) {
            NetworkWriter.print("Y");
          }
          else if (link.fLaneChangePermissions[i] == Link.kNotAllowed) {
            NetworkWriter.print("N");
          }
          if (i < (link.fNrOfLaneChangePermissions - 1)) {
            NetworkWriter.print("  ");
          }
        }
        NetworkWriter.println();
      }
      NetworkWriter.println("}");

      NetworkWriter.println();
      // save the environment-section
      NetworkWriter.println("[Environment]");
      NetworkWriter.println("# A legal environment is defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   type  parameters  [name]");
      NetworkWriter.println("#");
      NetworkWriter.println("# Following types (with their parameters) are supported");
      NetworkWriter.println("# (all coordinates are expressed in meters (integers)) :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   rectangle (parameters : color  fill  upper-left-node  lower-right-node)");
      NetworkWriter.println("#   polygon   (parameters : color  fill  #nodes  node(1)  ...  node(n))");
      NetworkWriter.println("#   oval      (parameters : color  fill  center-node  x-radius  y-radius)");
      NetworkWriter.println("#   railroad  (parameters : #nodes  node(1)  ...  node(n))");
      NetworkWriter.println("#");
      NetworkWriter.println("# where x-radius and y-radius are expressed in meters (integers)");
      NetworkWriter.println("# and trackwidth is expressed in meters (float).");
      NetworkWriter.println("#");
      NetworkWriter.println("# Available colors are : black, blue, cyan, gray, green, ");
      NetworkWriter.println("#                        magenta, orange, red, white, yellow.");
      NetworkWriter.println("# Available fills are : outlined, filled.");
      NetworkWriter.println("{");
      Enumeration EnvironmentEnum = fEnvironment.elements();
      while (EnvironmentEnum.hasMoreElements()) {
        Environment environment = (Environment) EnvironmentEnum.nextElement();

        NetworkWriter.print("  ");
        if (environment.fType == Environment.kRectangle) {
          NetworkWriter.print(expand("rectangle",10));
        }
        else if (environment.fType == Environment.kPolygon) {
          NetworkWriter.print(expand("polygon",10));
        }
        else if (environment.fType == Environment.kOval) {
          NetworkWriter.print(expand("oval",10));
        }
        else if (environment.fType == Environment.kRailroad) {
          NetworkWriter.print(expand("railroad",10));
        }
        NetworkWriter.print("  ");
        if ((environment.fType == Environment.kRectangle) ||
            (environment.fType == Environment.kPolygon) ||
            (environment.fType == Environment.kOval)) {
          if (environment.fColor == Color.black) {
            NetworkWriter.print(expand("black",8));
          }
          else if (environment.fColor == Color.blue) {
            NetworkWriter.print(expand("blue",8));
          }
          else if (environment.fColor == Color.cyan) {
            NetworkWriter.print(expand("cyan",8));
          }
          else if (environment.fColor == Color.gray) {
            NetworkWriter.print(expand("gray",8));
          }
          else if (environment.fColor == Color.green) {
            NetworkWriter.print(expand("green",8));
          }
          else if (environment.fColor == Color.magenta) {
            NetworkWriter.print(expand("magenta",8));
          }
          else if (environment.fColor == Color.orange) {
            NetworkWriter.print(expand("orange",8));
          }
          else if (environment.fColor == Color.red) {
            NetworkWriter.print(expand("red",8));
          }
          else if (environment.fColor == Color.white) {
            NetworkWriter.print(expand("white",8));
          }
          else if (environment.fColor == Color.yellow) {
            NetworkWriter.print(expand("yellow",8));
          }
          NetworkWriter.print("  ");
          if (environment.fFilled) {
            NetworkWriter.print(expand("filled",9));
          }
          else {
            NetworkWriter.print(expand("outlined",9));
          }
          NetworkWriter.print("  ");
        }
        if ((environment.fType == Environment.kPolygon) ||
            (environment.fType == Environment.kRailroad)) {
          NetworkWriter.print(expand(String.valueOf(environment.fNrOfNodes),9));
          NetworkWriter.print("  ");
        }
        for (int i = 0; i < environment.fNrOfNodes; ++i) {
          NetworkWriter.print(expand(String.valueOf(environment.fNodes[i].fID),kIndexWidth));
          if (i != (environment.fNrOfNodes - 1)) {
            NetworkWriter.print("  ");
          }
        }
        if (environment.fType == Environment.kOval) {
          NetworkWriter.print("  ");
          NetworkWriter.print(expand(String.valueOf(environment.fXRadius),kCoordinateWidth));
          NetworkWriter.print("  ");
          NetworkWriter.print(expand(String.valueOf(environment.fYRadius),kCoordinateWidth));
        }
        if (!environment.fName.equalsIgnoreCase("")) {
          NetworkWriter.print("  ");
          NetworkWriter.print("\"" + environment.fName + "\"");
        }
        NetworkWriter.println();
      }
      NetworkWriter.println("}");

      NetworkWriter.println();
      // save the model-section
      NetworkWriter.println("[Model]");
      NetworkWriter.println("# This section is defined as :");
      NetworkWriter.println("#");
      NetworkWriter.println("#   origin-xoffset  origin-yoffset  scalefactor  zoomfactor  map-xoffset  map-yoffset  [mapfilename]");
      NetworkWriter.println("{");
      NetworkWriter.println("  " +
                            String.valueOf(fOriginXOffset) +
                            "  " +
                            String.valueOf(fOriginYOffset) +
                            "  " +
                            String.valueOf(fScaleFactor) +
                            "  " +
                            String.valueOf(fZoomFactor) +
                            "  " +
                            String.valueOf(fMapXOffset) +
                            "  " +
                            String.valueOf(fMapYOffset) +
                            "  " +
                            "\"" + String.valueOf(fMapFilename) + "\"");
      NetworkWriter.println("}");

      NetworkWriter.close();
      return true;
    }
    catch (IOException exc) {
      return false;
    }
  }

  private String expand(String string, int width)
  {
    String Buffer = string;
    for (int i = 0; i < (width - string.length()); ++i) {
      Buffer = Buffer.concat(" ");
    }
    return Buffer;
  }

  public void translate(int dx, int dy, int dz)
  {
    Enumeration NodeEnum = fNodes.elements();
    while (NodeEnum.hasMoreElements()) {
      Node node = (Node) NodeEnum.nextElement();

      // translate the node
      node.fXPos += dx;
      node.fYPos += dy;
      node.fZPos += dz;
    }
  }

  public void readFromStream(DataInputStream in, boolean ShowTickMarks)
  {
    final int NrOfDotMarks = 5;

    try {
      clear();

      // handle the general settings
      fLaneWidth = in.readFloat();
      fRailroadWidth = in.readFloat();
      fTitle = in.readUTF();

      // handle the nodes
      int NrOfNodes = in.readInt();
      int CurrentNode = 0;
      int TickMark = NrOfNodes / NrOfDotMarks;
      for (int i = 0; i < NrOfNodes; ++i) {
        Node node = new Node();
        node.fID = in.readInt();
        node.fXPos = in.readInt();
        node.fYPos = in.readInt();
        node.fZPos = in.readInt();
        node.fType = in.readShort();
        node.fName = in.readUTF();
        fNodes.put(new Integer(node.fID),node);
        ++CurrentNode;
        if (ShowTickMarks && (TickMark != 0)) {
          if ((CurrentNode % TickMark) == 0) {
            System.out.print(".");
          }
        }
      }

      // handle the links
      int NrOfLinks = in.readInt();
      int CurrentLink = 0;
      TickMark = NrOfNodes / NrOfDotMarks;
      for (int i = 0; i < NrOfLinks; ++i) {
        Link link = new Link();
        link.fID = in.readInt();
        link.fNrOfNodes = in.readInt();
        link.fNodes = new Node[link.fNrOfNodes];
        for (int j = 0; j < link.fNrOfNodes; ++j) {
          int NodeID = in.readInt();
          link.fNodes[j] = (Node) (fNodes.get(new Integer(NodeID)));
        }
        link.fNrOfLanes = in.readInt();
        link.fLengths = new int[link.fNrOfNodes - 1];
        for (int j = 0; j < (link.fNrOfNodes - 1); ++j) {
          link.fLengths[j] = in.readInt();
        }
        link.fGrades = new float[link.fNrOfNodes - 1];
        for (int j = 0; j < (link.fNrOfNodes - 1); ++j) {
          link.fGrades[j] = in.readFloat();
        }
        link.fSpeedLimit = in.readInt();
        link.fIsTerminator = in.readBoolean();
        link.fMainLinkConnectedToLink = in.readInt();
        link.fMainLinkConnectedToPart = in.readInt();
        link.fType = in.readShort();
        link.fLocation = in.readShort();
        link.fOffRampConnectedToLink = in.readInt();
        link.fOffRampConnectedToPart = in.readInt();
        link.fNrOfLanesOnRamp = in.readInt();
        link.fNewNrOfLanes = in.readInt();
        link.fNrOfLaneChangePermissions = in.readInt();
        link.fLaneChangePermissions = new int[link.fNrOfLaneChangePermissions];
        for (int j = 0; j < link.fNrOfLaneChangePermissions; ++j) {
          link.fLaneChangePermissions[j] = in.readInt();
        }
        link.fName = in.readUTF();
        link.fRampLength = in.readFloat();
        fLinks.put(new Integer(link.fID),link);
        ++CurrentLink;
        if (ShowTickMarks && (TickMark != 0)) {
          if ((CurrentLink % TickMark) == 0) {
            System.out.print(".");
          }
        }
      }

      // handle the environment
      int NrOfEnvironments = in.readInt();
      int CurrentEnvironment = 0;
      TickMark = NrOfNodes / NrOfDotMarks;
      for (int i = 0; i < NrOfEnvironments; ++i) {
        Environment environment = new Environment();

        environment.fType = in.readShort();
        int Red = in.readInt();
        int Green = in.readInt();
        int Blue = in.readInt();
        environment.fColor = new Color(Red,Green,Blue);
        environment.fFilled = in.readBoolean();
        environment.fNrOfNodes = in.readInt();
        environment.fNodes = new Node[environment.fNrOfNodes];
        for (int j = 0; j < environment.fNrOfNodes; ++j) {
          int NodeID = in.readInt();
          environment.fNodes[j] = (Node) (fNodes.get(new Integer(NodeID)));
        }
        environment.fXRadius = in.readInt();
        environment.fYRadius = in.readInt();
        environment.fName = in.readUTF();
        environment.calcBoundingBox();
        fEnvironment.addElement(environment);
        ++CurrentEnvironment;
        if (ShowTickMarks && (TickMark != 0)) {
          if ((CurrentEnvironment % TickMark) == 0) {
            System.out.print(".");
          }
        }
      }
    }
    catch (IOException exc) {
    }
  }

  public void writeToStream(DataOutputStream out)
  {
    try {
      // handle the general settings
      out.writeFloat(fLaneWidth);
      out.writeFloat(fRailroadWidth);
      out.writeUTF(fTitle);

      // handle the nodes
      out.writeInt(fNodes.size());
      Enumeration NodeEnum = fNodes.elements();
      while (NodeEnum.hasMoreElements()) {
        Node node = (Node) NodeEnum.nextElement();

        out.writeInt(node.fID);
        out.writeInt(node.fXPos);
        out.writeInt(node.fYPos);
        out.writeInt(node.fZPos);
        out.writeShort(node.fType);
        out.writeUTF(node.fName);
      }
      out.flush();

      // handle the links
      out.writeInt(fLinks.size());
      Enumeration LinkEnum = fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        out.writeInt(link.fID);
        out.writeInt(link.fNrOfNodes);
        for (int i = 0; i < link.fNrOfNodes; ++i) {
          out.writeInt(link.fNodes[i].fID);
        }
        out.writeInt(link.fNrOfLanes);
        for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
          out.writeInt(link.fLengths[i]);
        }
        for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
          out.writeFloat(link.fGrades[i]);
        }
        out.writeInt(link.fSpeedLimit);
        out.writeBoolean(link.fIsTerminator);
        out.writeInt(link.fMainLinkConnectedToLink);
        out.writeInt(link.fMainLinkConnectedToPart);
        out.writeShort(link.fType);
        out.writeShort(link.fLocation);
        out.writeInt(link.fOffRampConnectedToLink);
        out.writeInt(link.fOffRampConnectedToPart);
        out.writeInt(link.fNrOfLanesOnRamp);
        out.writeInt(link.fNewNrOfLanes);
        out.writeInt(link.fNrOfLaneChangePermissions);
        for (int i = 0; i < link.fNrOfLaneChangePermissions; ++i) {
          out.writeInt(link.fLaneChangePermissions[i]);
        }
        out.writeUTF(link.fName);
        out.writeFloat(link.fRampLength);
      }
      out.flush();

      // handle the environment
      out.writeInt(fEnvironment.size());
      Enumeration EnvironmentEnum = fEnvironment.elements();
      while (EnvironmentEnum.hasMoreElements()) {
        Environment environment = (Environment) EnvironmentEnum.nextElement();

        out.writeShort(environment.fType);
        out.writeInt(environment.fColor.getRed());
        out.writeInt(environment.fColor.getGreen());
        out.writeInt(environment.fColor.getBlue());
        out.writeBoolean(environment.fFilled);
        out.writeInt(environment.fNrOfNodes);
        for (int i = 0; i < environment.fNrOfNodes; ++i) {
          out.writeInt(environment.fNodes[i].fID);
        }
        out.writeInt(environment.fXRadius);
        out.writeInt(environment.fYRadius);
        out.writeUTF(environment.fName);
      }

      out.flush();
    }
    catch (IOException exc) {
    }
  }

  public void readTimeStampFromStream(DataInputStream in)
  {
    try {
      fTimeStamp = in.readLong();
    }
    catch (IOException exc) {
    }
  }

  public void writeTimeStampToStream(DataOutputStream out)
  {
    try {
      out.writeLong(fTimeStamp);
      out.flush();
    }
    catch (IOException exc) {
    }
  }
}
